"""Mount broker."""
from . import Broker, python3, load_output
from jujuna.exporters import Exporter
import re
import logging


log = logging.getLogger('jujuna.tests.broker')


class Mount(Broker):
    """Mount broker."""

    def __init__(self):
        """Init broker."""
        super().__init__()

    async def run(self, test_case, unit, idx):
        """Run tests."""
        rows = []
        async with Exporter(unit, self.named) as exporter:
            try:
                act = await unit.run(python3(exporter), timeout=10)
                results = load_output(act.data['results'])
            except Exception as exc:
                log.debug(exc)
                results = []
            # print(results.keys())
            if 'regex' in test_case:
                for condition in test_case['regex']:
                    prog = re.compile(condition)
                    mounts = ''

                    for result in results:
                        var = prog.search(result)
                        if var:
                            mounts = var.group(0)

                    rows.append((idx, '{} == {}'.format(
                        mounts if mounts else str(condition), 'mounted'
                    ), True if mounts else False), )

        return rows



#!/usr/bin/python3

import sys
import logging
import argparse
import argcomplete
import async_timeout

from juju import loop

from jujuna.helper import log_traceback

from jujuna.deploy import deploy  # noqa
from jujuna.upgrade import upgrade  # noqa
from jujuna.tests import test  # noqa
from jujuna.clean import clean  # noqa


logger = logging.getLogger('jujuna')
logger.setLevel(logging.INFO)

# Double logging cleanup
while logger.handlers:
    logger.handlers.pop()

logHandler = logging.StreamHandler(sys.stdout)
logHandler.setLevel(logging.INFO)
logFormatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logHandler.setFormatter(logFormatter)
logger.addHandler(logHandler)


def get_parser():
    parser = argparse.ArgumentParser(
        description="Deploy a local bundle, execute upgrade procedure, "
                    "run the deployment through a suite of tests to ensure "
                    "that it can handle the types of operations and failures "
                    "that are common for all deployments.",
    )

    subparsers = parser.add_subparsers(dest='action', help='Action to be executed')
    subparsers.required = True

    p_deploy = subparsers.add_parser('deploy', help="Deploy a local bundle to the current or selected model")
    p_deploy.add_argument('bundle_file', type=argparse.FileType('r'),
                          help="Path to bundle file (i.e. bundle.yaml)")
    p_deploy.add_argument("-c", "--controller", default=None, dest="ctrl_name", help="Controller (def: current)")
    p_deploy.add_argument("-m", "--model", default=None, dest="model_name", help="Model to use instead of current")
    p_deploy.add_argument("-w", "--wait", action='store_true', help="Wait for deploy to finish")
    p_deploy.add_argument("-t", "--timeout", default=0, type=int, help="Timeout after N seconds.")
    p_deploy.add_argument("--error-timeout", default=1800, dest="error_timeout", type=int,
                          help="Timeout after N seconds in error state.")
    p_deploy.add_argument("--endpoint", default=None, dest="endpoint",
                          help="Juju endpoint (requires model uuid instead of name)")
    p_deploy.add_argument("--username", default=None, dest="username", help="Juju username")
    p_deploy.add_argument("--password", default=None, dest="password", help="Juju password")
    p_deploy.add_argument("--cacert", default=None, dest="cacert", help="Juju CA certificate")
    p_deploy.add_argument("--debug", action='store_true', help="Log level debug.")

    p_upgrade = subparsers.add_parser('upgrade', help="Upgrade applications deployed in the current or selected model")
    p_upgrade.add_argument("-c", "--controller", default=None, dest="ctrl_name", help="Controller (def: current)")
    p_upgrade.add_argument("-m", "--model", default=None, dest="model_name", help="Model to use instead of current")
    p_upgrade.add_argument("-o", "--origin", default='', dest="origin",
                           help="""Openstack origin:
                                   'cloud:xenial-newton',
                                   'cloud:xenial-ocata',
                                   'cloud:xenial-pike',
                                   'cloud:xenial-queens',
                                   'cloud:bionic-rocky',
                                   'cloud:bionic-stein',
                                   'cloud:bionic-train',
                                """)
    p_upgrade.add_argument("-a", "--apps", nargs='*', default=[], help="Apps to be upgraded (ordered)")
    p_upgrade.add_argument("-i", "--ignore-errors", action='store_true', dest='ignore_errors',
                           help="Ignore errors during charms upgrade and continue with upgrade procedure")
    p_upgrade.add_argument("-p", "--pause", action='store_true', help="Pause unit before upgrade (incl. HA)")
    p_upgrade.add_argument("-e", "--evacuate", action='store_true', help="Evacuate nova-compute nodes during upgrade")
    p_upgrade.add_argument("--upgrade-only", action='store_true', dest="upgrade_only",
                           help="Upgrade using upgrade hooks without changing the revision")
    p_upgrade.add_argument("--charms-only", action='store_true', dest="charms_only",
                           help="Upgrade only charms without running upgrade hooks")
    p_upgrade.add_argument("--upgrade-action", dest="upgrade_action", default=None,
                           help="Action name to upgrade application")
    p_upgrade.add_argument("--upgrade-params", dest="upgrade_params", default=None,
                           help="Action parameters comma separated e.g. 'service=name,version=2'")
    p_upgrade.add_argument("--origin-keys", dest="origin_keys", default=None,
                           help="Config keys to set origin in apps e.g. 'ceph-mon=source,ceph-mon=source'")
    p_upgrade.add_argument("--dry-run", action='store_true', dest="dry_run",
                           help="Dry run - only show changes without upgrading")
    p_upgrade.add_argument("-t", "--timeout", default=0, type=int, help="Timeout after N seconds.")
    p_upgrade.add_argument("-s", "--settings", type=argparse.FileType('r'),
                           help="Path to settings file that overrides default settings (i.e. settings.yaml)")
    p_upgrade.add_argument("--endpoint", default=None, dest="endpoint",
                           help="Juju endpoint (requires model uuid instead of name)")
    p_upgrade.add_argument("--username", default=None, dest="username", help="Juju username")
    p_upgrade.add_argument("--password", default=None, dest="password", help="Juju password")
    p_upgrade.add_argument("--cacert", default=None, dest="cacert", help="Juju CA certificate")
    p_upgrade.add_argument("--debug", action='store_true', help="Log level debug.")

    p_test = subparsers.add_parser('test', help="Test applications in the current or selected model")
    p_test.add_argument("test_suite", type=argparse.FileType('r'),
                        help="Path to test suite (i.e. ceph/suite.yaml)")
    p_test.add_argument("-c", "--controller", default=None, dest="ctrl_name", help="Controller (def: current)")
    p_test.add_argument("-m", "--model", default=None, dest="model_name", help="Model to use instead of current")
    p_test.add_argument("-t", "--timeout", default=0, type=int, help="Timeout after N seconds.")
    p_test.add_argument("--endpoint", default=None, dest="endpoint",
                        help="Juju endpoint (requires model uuid instead of name)")
    p_test.add_argument("--username", default=None, dest="username", help="Juju username")
    p_test.add_argument("--password", default=None, dest="password", help="Juju password")
    p_test.add_argument("--cacert", default=None, dest="cacert", help="Juju CA certificate")
    p_test.add_argument("--debug", action='store_true', help="Log level debug.")

    p_clean = subparsers.add_parser(
        'clean',
        help="Clean the model by removing all applications present in the current or selected model"
    )
    p_clean.add_argument("-c", "--controller", default=None, dest="ctrl_name", help="Controller (def: current)")
    p_clean.add_argument("-m", "--model", default=None, dest="model_name", help="Model to use instead of current")
    p_clean.add_argument("-w", "--wait", action='store_true', help="Wait for deploy to finish")
    p_clean.add_argument("-f", "--force", action='store_true', help="Force cleanup (remove all machines in the model).")
    p_clean.add_argument("-i", "--ignore", nargs='*', default=[], help="Apps to be ignored during removal")
    p_clean.add_argument("--dry-run", action='store_true', dest="dry_run",
                         help="Dry run - only show changes without removing applications")
    p_clean.add_argument("-t", "--timeout", default=0, type=int, help="Timeout after N seconds.")
    p_clean.add_argument("--endpoint", default=None, dest="endpoint",
                         help="Juju endpoint (requires model uuid instead of name)")
    p_clean.add_argument("--username", default=None, dest="username", help="Juju username")
    p_clean.add_argument("--password", default=None, dest="password", help="Juju password")
    p_clean.add_argument("--cacert", default=None, dest="cacert", help="Juju CA certificate")
    p_clean.add_argument("--debug", action='store_true', help="Log level debug.")

    argcomplete.autocomplete(parser)

    return parser


doc_arg_parser = get_parser()  # To be included in documentation


async def run_action(action, timeout, args):
    """Run request action.

    """
    selected_action = globals()[action]

    if timeout == 0:
        return await selected_action(**args)
    else:
        at = None
        try:
            async with async_timeout.timeout(timeout) as at:
                ret = await selected_action(**args)
        except Exception as e:
            # Exit with timeout code if expired
            if at and at.expired:
                ret = 124
                logger.warn('Action {} timed out!'.format(action))
            else:
                ret = 1
                logger.warn('Action {} failed!'.format(action))
                logger.warn(str(e))
        return ret


def load_kv_arg(args, item):
    if args.get(item, False):
        up_list = args[item].split(',')
        upgrade_params = dict([
            item.split('=') if '=' in item else [item, True] for item in up_list
        ])
        args[item] = upgrade_params
    else:
        args[item] = {}
    return args


def parse_args(argv):
    parser = get_parser()
    try:
        parsed = parser.parse_args(argv)
        args = vars(parsed)
        action = args.pop('action')
        timeout = args.pop('timeout')

        args = load_kv_arg(args, 'upgrade_params')
        args = load_kv_arg(args, 'origin_keys')
    except Exception as e:
        log_traceback(e)
        parser.print_help()
        sys.exit(1)
    try:
        if args.get('debug', False):
            logger.setLevel(logging.DEBUG)
            logHandler.setLevel(logging.DEBUG)
    except Exception:
        pass
    return action, timeout, args


def main():
    action, timeout, args = parse_args(sys.argv[1:])

    try:
        ret = loop.run(
            run_action(action, timeout, args)
        )
    except Exception as e:
        log_traceback(e)
        ret = 1

    sys.exit(ret)


if __name__ == '__main__':
    main()



#!/usr/bin/python3

import logging
from jujuna.helper import connect_juju, log_traceback, ApperrorTimeout, wait_until
from juju.errors import JujuError
from websockets import ConnectionClosed


# create logger
log = logging.getLogger('jujuna.deploy')


async def deploy(
    bundle_file,
    ctrl_name='',
    model_name='',
    wait=False,
    endpoint='',
    username='',
    password='',
    cacert='',
    error_timeout=None,
    **kwargs
):
    """Deploy a local juju bundle.

    Handles deployment of a bundle file to the current or selected model.

    Connection requires juju client configs to be present locally or specification of credentialls:
    endpoint (e.g. 127.0.0.1:17070), username, password, and model uuid as model_name.

    :param bundle_file: juju bundle file
    :param ctrl_name: juju controller
    :param model_name: juju model name or uuid
    :param wait: boolean
    :param endpoint: string
    :param username: string
    :param password: string
    :param cacert: string
    """
    ret = 0
    log.info('Reading bundle: {}'.format(bundle_file.name))
    entity_url = 'local:' + bundle_file.name.replace('/bundle.yaml', '')

    controller, model = await connect_juju(
        ctrl_name,
        model_name,
        endpoint=endpoint,
        username=username,
        password=password,
        cacert=cacert
    )

    try:
        # Deploy a bundle
        log.info("Deploy: {}".format(entity_url))
        deployed_apps = await model.deploy(
            entity_url
        )
        if not isinstance(deployed_apps, list):
            deployed_apps = [deployed_apps]

        if wait:
            await wait_until(
                model,
                deployed_apps,
                log,
                loop=model.loop,
                error_timeout=error_timeout
            )
        else:
            log.info('{} - Apps: {}'.format(
                'DEPLOYED',
                len(deployed_apps)
            ))

    except ApperrorTimeout:
        ret = 124
        log.error('FAILED - Application too long in error state')
    except ConnectionClosed as e:
        ret = 1
        log.error('FAILED - Connection closed')
        log_traceback(e)
    except JujuError as e:
        ret = 1
        log.error('JujuError during deploy')
        log_traceback(e)
    finally:
        # Disconnect from the api server and cleanup.
        await model.disconnect()
        await controller.disconnect()
    return ret



"""Process broker."""
from . import Broker, python3, load_output
from jujuna.exporters import Exporter
import logging


log = logging.getLogger('jujuna.tests.broker')


class Process(Broker):
    """Process broker."""

    def __init__(self):
        """Init Process broker."""
        super().__init__()

    async def run(self, test_case, unit, idx):
        """Run tests."""
        rows = []
        async with Exporter(unit, self.named) as exporter:
            try:
                act = await unit.run(python3(exporter), timeout=10)
                results = load_output(act.data['results'])
            except Exception as exc:
                log.debug(exc)
                results = {}
            # print(results)
            for condition, present in test_case.items():
                status = 'present' if present else 'absent'
                rows.append((idx, '{} == {}'.format(condition, status), (condition in results) == present), )

        return rows



#!/usr/bin/env python3

import json
import sys


def main():
    mount_data = {}
    with open('/proc/mounts', 'r') as mountfile:
        mount_lines = mountfile.read().split('\n')[:-1]
        for mount_line in mount_lines:
            mount_items = mount_line.split(' ')
            mount_parsed = {
                'mountpoint': mount_items[1],
                'fs': mount_items[2],
                'params': mount_items[3].split(','),
            }
            if '/' not in mount_items[0]:
                if mount_items[0] not in mount_data:
                    mount_data[mount_items[0]] = []
                mount_data[mount_items[0]].append(mount_parsed)
            else:
                mount_data[mount_items[0]] = mount_parsed

    print(json.dumps(mount_data))
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
import json
import sys


def main():
    pids = [pid for pid in os.listdir('/proc') if pid.isdigit()]
    pnames = {}
    for pid in pids:
        pname_raw = open(os.path.join('/proc', pid, 'cmdline'), 'rb').read()
        pname = pname_raw.decode('utf-8').replace('\x00', ' ').strip().split(' ')
        pnames[pname[0]] = {
            'pid': pid,
            'params': pname[1:] if len(pname) > 1 else [],
        }

    print(json.dumps(pnames))
    sys.exit(0)


if __name__ == "__main__":
    main()



"""File broker."""
from . import Broker, python3, load_output
from jujuna.exporters import Exporter
import logging


log = logging.getLogger('jujuna.tests.broker')


class File(Broker):
    """File broker."""

    def __init__(self):
        """Init broker."""
        super().__init__()

    async def run(self, test_data, unit, idx):
        """Run tests."""
        rows = []
        async with Exporter(unit, self.named) as exporter:
            files = test_data
            # print(test_data)

            for file, params in files.items():
                try:
                    act = await unit.run(python3(exporter, args=[file]), timeout=10)
                    results = load_output(act.data['results'])
                except Exception as exc:
                    log.debug(exc)
                    results = {}
                # print('Expect: ', files[file])
                # print(results)
                for param, value in params.items():
                    res = (param in results) and (results[param] == value)
                    rows.append((idx, '{}.{} == {}'.format(file, param, value), res), )

        return rows



"""Exporter management."""

import os
import uuid
import logging
import async_timeout


log = logging.getLogger('jujuna.tests.exporter')


class Exporter():
    """Exporter context manager."""

    def __init__(self, unit, exporter_name, timeout=15):
        """Init exporter."""
        if exporter_name.endswith('.py'):
            raise Exception('Do not provide an extension')
        self.timeout = timeout
        self.unit = unit
        self.full_path = os.path.dirname(os.path.realpath(__file__))
        self.exporter_local = os.path.join(self.full_path, exporter_name + '.py')
        if not os.path.isfile(self.exporter_local):
            raise Exception('Exporter: {}.py not found'.format(exporter_name))
        self.exporter_remote = os.path.join('/tmp', str(uuid.uuid4())[:9] + exporter_name + '.py')

    async def __aenter__(self):
        """Upload exporter and return remote path."""
        # print('Load: ', self.exporter_local)
        await self._load_exporter(self.unit, self.exporter_local, self.exporter_remote)
        return self.exporter_remote

    async def __aexit__(self, type, value, traceback):
        """Dispose remote exporter."""
        # print('Unload: ', self.exporter_remote)
        await self._unload_exporter(self.unit, self.exporter_remote)

    async def _load_exporter(self, unit, source, target, user='ubuntu'):
        """Upload exporter."""
        for i in range(3):
            try:
                async with async_timeout.timeout(self.timeout):
                    await unit.scp_to(
                        source,
                        target,
                        user=user
                    )
                break
            except Exception as exc:
                log.debug(exc)

    async def _unload_exporter(self, unit, target, user='ubuntu'):
        """Dispose of remote exporter."""
        ret = None
        for i in range(3):
            try:
                async with async_timeout.timeout(self.timeout):
                    ret = await unit.run('rm {}'.format(target), timeout=self.timeout - 1)
                break
            except Exception as exc:
                # Try 3 times, if it fails raise on last
                if i == 2 and 'No such file' not in str(exc):
                    log.warning('Unload failed: {}'.format(str(exc)))
                    log.debug(exc)
        # ret = await unit.run('rm /tmp/*.py', timeout=3)
        if (
            ret and ret.data and
            'results' in ret.data and
            ret.data['results']['Code'] != '0' and
            'No such file' not in ret.data['results']['Stderr']
        ):
            log.warn('Unload failed: {}'.format(ret.data['results']['Stderr']))



#!/usr/bin/env python3

import re
import sys
import json
import base64
import ipaddress


def decode_address(address):
    hex_ip, hex_port = address.split(':')
    ip_join, ipv6 = (':', True) if len(hex_ip) == 32 else ('.', False)
    if ipv6:
        ipv6_string = ":".join([hex_ip[i:i + 4] for i in range(0, 32, 4)])
        ip = ipaddress.ip_address(ipv6_string).exploded
    else:
        ip_items = [str(item) for item in reversed(list(base64.b16decode(hex_ip)))]
        ip = ip_join.join(ip_items)
    port = int.from_bytes(base64.b16decode(hex_port), byteorder='big')
    return ip, port


def _read_tcp(path='/proc/net/tcp'):
    lines = []
    with open(path, 'r') as f:
        for line in f:
            line = line.strip()
            line = re.sub(' +', ' ', line)
            lines.append(line.split())
    sock_list = []
    header = lines[0]
    for line in lines[1:]:
        line_struct = {}
        for idx, head in enumerate(header):
            key = 'remote_address' if head == 'rem_address' else head
            line_struct[key] = line[idx]
        sock_list.append(line_struct)
    return sock_list


def sockets():
    sock_list = []
    sock_list.extend(_read_tcp())
    sock_list.extend(_read_tcp('/proc/net/tcp6'))

    data_list = []
    for item in sock_list:
        data_item = {}
        data_item['local_ip'], data_item['local_port'] = decode_address(item['local_address'])
        data_item['rem_ip'], data_item['rem_port'] = decode_address(item['remote_address'])
        data_list.append(data_item)

    return data_list


def main():
    data = {
        'sockets': sockets(),
    }

    print(json.dumps(data))
    sys.exit(0)


if __name__ == "__main__":
    main()



import asyncio
import logging
import traceback
from collections import Counter
from websockets import ConnectionClosed

from jujuna.settings import MAX_FRAME_SIZE

from juju.controller import Controller
from juju.model import Model

from theblues.charmstore import CharmStore


class ApperrorTimeout(Exception):
    """Raised when an application stays too long in error state."""

    def __init__(self):
        self.message = 'Timed out with application in error state'


def log_traceback(ex, prefix=''):
    if prefix and isinstance(prefix, str):
        prefix = '{} - '.format(prefix.strip())
    ex_traceback = ex.__traceback__
    for line in traceback.format_exception(
        ex.__class__, ex, ex_traceback
    ):
        logging.error('{}{}'.format(prefix, line.rstrip('\n')))


def cs_name_parse(name, series=None):
    """Charm store name parse.

    Parse charm name, series, and revision.
    Charmstore variable differs source from local charm.

    :param name string: charm name or url e.g. cs:xenial/glance-35
    :param series string: series name e.g. bionic
    :return dict: with keys series, charm, revision, and charmstore
    """
    try:
        cscode, csuri = name.split(':', 1) if ':' in name else ['cs', name]
        is_path = (cscode == 'local') or csuri.startswith('/') or csuri.startswith('./')
        uri_series, csuri = csuri.split('/', 1) if (
            '/' in csuri and not is_path
        ) else ['', csuri]
        items = csuri.split('-')
        revision = int(items[-1]) if items[-1].isdigit() else None
        return {
            'series': uri_series if uri_series and not series else series,
            'charm': '-'.join(items if revision is None else items[:-1]),
            'revision': revision,
            'charmstore': False if is_path else True
        }
    except Exception:
        logging.warning('Failed to parse Charm name: {}'.format(name))
        return {
            'series': None,
            'charm': name,
            'revision': None,
            'charmstore': False
        }


async def connect_juju(ctrl_name=None, model_name=None, endpoint=None, username=None, password=None, cacert=None):
    controller = Controller(max_frame_size=MAX_FRAME_SIZE)  # noqa

    if endpoint:
        await controller.connect(endpoint=endpoint, username=username, password=password, cacert=cacert)
    else:
        await controller.connect(ctrl_name)

    if endpoint:
        model = Model(max_frame_size=MAX_FRAME_SIZE)
        await model.connect(uuid=model_name, endpoint=endpoint, username=username, password=password, cacert=cacert)
    elif model_name:
        model = await controller.get_model(model_name)
    else:
        model = Model(max_frame_size=MAX_FRAME_SIZE)  # noqa
        await model.connect()

    # HACK low unsettable timeout in the model
    model.charmstore._cs = CharmStore(timeout=60)

    return controller, model


async def wait_until(model, apps, logger, log_time=5, timeout=None, wait_period=0.5, error_timeout=None, loop=None):
    """Blocking with log messages.

    Return only after all conditions are true.

    :param model: juju model
    :param apps: juju application
    :param logger: logger instance
    :param log_time: logging frequency (s)
    :param timeout: blocking timeout (s)
    :param error_timeout: timeout if app persists in error state for set time (s)
    :param wait_period: waiting time between checks (s)
    :param loop: asyncio event loop
    """
    log_count = 0

    def _disconnected():
        return not (model.is_connected() and model.connection().is_open)

    async def _block(log_count):
        blockable = ['maintenance', 'blocked', 'waiting', 'error']
        success_counter = 0
        errors = 0
        # 10 consecutive success checks to pass
        while success_counter < 5:
            success_counter += 1
            while (
                not _disconnected() and
                (
                    any(a.status != 'active' for a in apps) or
                    any(u.workload_status in blockable for a in apps for u in a.units)
                )
            ):
                success_counter = 0  # checks not successful - start again at 0
                log_count += 0.5
                if any(a.status == 'error' for a in apps):
                    errors += wait_period
                if error_timeout and errors >= max(error_timeout, 20):
                    log_workload(
                        logger, model, apps,
                        label='FAILED', error_status=True, status=False, workload=False
                    )
                    raise ApperrorTimeout()
                await asyncio.sleep(wait_period, loop=loop)
                if log_count % log_time == 0:
                    log_workload(logger, model, apps)
            if success_counter != 0:
                # Log if increasing consecutive check count
                if all(a.status != 'error' for a in apps):
                    errors = 0
                await asyncio.sleep(2, loop=loop)
                log_workload(logger, model, apps)

    await asyncio.wait_for(_block(log_count), timeout, loop=loop)

    if _disconnected():
        raise ConnectionClosed(1006, 'no reason')
    else:
        log_workload(logger, model, apps, label='DEPLOYED')


def log_workload(
    logger, model, apps, label='PROGRESS', status=True, workload=True, error_status=False, unit_status=False
):
    if status:
        ass = dict(Counter([a.status if a.status else 'none' for a in apps]))
        ass = 'Status: {}'.format(','.join([
            '{}={}'.format(k, v) for k, v in ass.items()
        ])) if ass else 'Status: None'
    else:
        ass = ''
    if workload:
        wss = Counter([
            unit.workload_status if unit.workload_status else 'none' for a in apps for unit in a.units
        ])
        wss = 'Workload: {}'.format(','.join([
            '{}={}'.format(k, v) for k, v in wss.items()
        ])) if wss else 'Workload: None'
    else:
        wss = ''

    if error_status:
        asm = dict(Counter([a.status_message for a in apps if a.status == 'error']))
        asm = 'Errors: {}'.format(','.join(asm.keys())) if asm else 'Errors: None'
    else:
        asm = ''
    if unit_status:
        wsm = dict(Counter([
            (
                u.workload_status_message if 'ready' not in u.workload_status_message else 'ready'
            ) for a in apps for u in a.units
        ]))
        wsm = 'Units: {}'.format('{}={}'.format(k, v) for k, v in wsm.items()) if wsm else 'Units: None'
    else:
        wsm = ''
    msg = ' '.join([x for x in [ass, wss, asm, wsm] if x])
    logger.info('{} - Machines {} - Apps {} {}'.format(
        label,
        len(model.machines),
        len(model.applications),
        '- {}'.format(msg) if msg else ''
    ))



#!/usr/bin/env python3

import json
import sys
import os
import stat


def main():
    if len(sys.argv) != 2:
        raise Exception('Specify exactly one filepath.')
    filepath = sys.argv[1]
    if not os.path.exists(filepath):
        raise Exception("Path '{}' does not exist.".format(filepath))

    file_stat = os.stat(filepath)

    file_vars = {
        'st_mode': file_stat.st_mode,
        'st_ino': file_stat.st_ino,
        'st_dev': file_stat.st_dev,
        'st_nlink': file_stat.st_nlink,
        'st_uid': file_stat.st_uid,
        'st_gid': file_stat.st_gid,
        'st_size': file_stat.st_size,
        'st_atime': file_stat.st_atime,
        'st_mtime': file_stat.st_mtime,
        'st_ctime': file_stat.st_ctime,
        'is_dir': stat.S_ISDIR(file_stat.st_mode),
        'is_chr': stat.S_ISCHR(file_stat.st_mode),
        'is_blk': stat.S_ISBLK(file_stat.st_mode),
        'is_reg': stat.S_ISREG(file_stat.st_mode),
        'is_fifo': stat.S_ISFIFO(file_stat.st_mode),
        'is_lnk': stat.S_ISLNK(file_stat.st_mode),
        'is_sock': stat.S_ISSOCK(file_stat.st_mode),
        'imode': stat.S_IMODE(file_stat.st_mode),
        'ifmt': stat.S_IFMT(file_stat.st_mode),
    }

    print(json.dumps(file_vars))
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import json
import sys
import dbus


def main():
    bus = dbus.SystemBus()
    systemd1 = bus.get_object('org.freedesktop.systemd1', '/org/freedesktop/systemd1')
    manager = dbus.Interface(systemd1, dbus_interface='org.freedesktop.systemd1.Manager')

    units_list = manager.ListUnits()

    services = {}

    for unit in units_list:
        if unit[0].endswith('.service'):
            unit_name = unit[0].replace('.service', '')

            services[unit_name] = {
                'name': unit[0],
                'description': unit[1],
                'loaded': unit[2],
                'active': unit[3],
                'status': unit[4],
            }

    systemd_data = {
        'services': services,
    }

    print(json.dumps(systemd_data))
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import json
import sys
import pwd
import grp


def main():

    user_data = {}

    for user in pwd.getpwall():
        user_data[user[0]] = {
            'uid': user[2],
            'gid': user[3],
            'group': grp.getgrgid(user[3])[0],
            'gecos': user[4],
            'dir': user[5],
            'shell': user[6],
        }

    print(json.dumps(user_data))
    sys.exit(0)


if __name__ == "__main__":
    main()



import asyncio
import logging
import yaml
import websockets
import async_timeout

from collections import Counter

from jujuna.helper import cs_name_parse, connect_juju, log_traceback
from jujuna.settings import ORIGIN_KEYS, SERVICES

from juju.errors import JujuError


# create logger
log = logging.getLogger('jujuna.upgrade')


async def upgrade(
    ctrl_name=None,
    model_name=None,
    apps=[],
    origin='',
    ignore_errors=False,
    pause=False,
    evacuate=False,
    charms_only=False,
    upgrade_only=False,
    upgrade_action='',
    upgrade_params={},
    origin_keys={},
    dry_run=False,
    settings=None,
    endpoint='',
    username='',
    password='',
    cacert='',
    **kwargs
):
    """Upgrade applications deployed in the model.

    Handles upgrade of application deployed in the specified model. Focused on openstack upgrade procedures.

    Connection requires juju client configs to be present locally or specification of credentialls:
    endpoint (e.g. 127.0.0.1:17070), username, password, and model uuid as model_name.

    :param ctrl_name: juju controller
    :param model_name: juju model name or uuid
    :param apps: ordered list of application names
    :param origin: target openstack version string e.g. 'cloud:xenial-ocata'
    :param ignore_errors: boolean
    :param pause: boolean
    :param evacuate: boolean
    :param charms_only: boolean
    :param upgrade_only: boolean
    :param upgrade_action: string
    :param upgrade_params: dict
    :param origin_keys: dict
    :param dry_run: boolean
    :param endpoint: string
    :param username: string
    :param password: string
    :param cacert: string
    """

    controller, model = await connect_juju(
        ctrl_name,
        model_name,
        endpoint=endpoint,
        username=username,
        password=password,
        cacert=cacert
    )

    try:
        log.info('Applications present in the current model: {}'.format(
            ', '.join(list(model.applications.keys()))
        ))

        settings_data = {}

        try:
            if settings:
                with open(settings.name, 'r') as stream:
                    if hasattr(yaml, 'full_load'):
                        settings_data = yaml.full_load(stream)
                    else:
                        settings_data = yaml.load(stream)
        except yaml.YAMLError as e:
            log.warn('Failed to load settings file: {}'.format(str(e)))

        origin_keys = settings_data.get('origin_keys', origin_keys if origin_keys else ORIGIN_KEYS)
        services = settings_data.get('services', SERVICES)
        add_services = settings_data.get('add_services', [])

        # If apps are not specified in the order use configuration from settings
        if apps:
            services = apps
            add_services = []

        log.info('Services to upgrade: {}'.format(', '.join(services)))
        if add_services and not upgrade_only:
            log.info('Charms only upgrade: {}'.format(', '.join(add_services)))
        all_services = [cs_name_parse(serv) for serv in services]
        applications = [serv['charm'] for serv in all_services]
        all_services.extend([cs_name_parse(serv) for serv in add_services])

        # Upgrade charm revisions
        if not upgrade_only:
            upgraded, latest_charms = await upgrade_charms(
                model, all_services, dry_run, ignore_errors
            )

        # Ocata upgrade requires additional relation to succeed
        if not charms_only and origin == 'cloud:xenial-ocata' and 'nova-compute' in applications:
            if 'cinder-warmceph' in applications:
                await ocata_relation_patch(model, dry_run, cinder_ceph='cinder-warmceph')
            elif 'cinder-ceph' in applications:
                await ocata_relation_patch(model, dry_run, cinder_ceph='cinder-ceph')

        # Upgrade applications
        if not charms_only:
            await upgrade_services(
                model, applications, origin, origin_keys, upgrade_action, upgrade_params, pause, dry_run
            )

        # Log status values
        d = Counter([a.status for a in model.applications.values()])
        log.info('STATUS: {}'.format(dict(d)))

    finally:
        await model.disconnect()
        await controller.disconnect()


async def ocata_relation_patch(model, dry_run, cinder_ceph):
    cinder_ceph_rel = '{}:ceph-access'.format(cinder_ceph)
    log.info('Adding relation between nova-compute:ceph-access and {}'.format(cinder_ceph_rel))
    if not dry_run:
        try:
            await model.add_relation('nova-compute:ceph-access', cinder_ceph_rel)
            # TODO add completion check
            await asyncio.sleep(120)
            await wait_until(
                model,
                model.applications.values(),
                timeout=1800,
                loop=model.loop
            )
            log.info('Completed addition of relation')
        except Exception as e:
            if 'already exists' not in str(e):
                log_traceback(e)
                raise e
            else:
                log.warn('Ignored: relation already exists')


def get_service_list(model, upgraded):
    return [(name, get_service_version(model.applications[name])) for name in upgraded if name in model.applications]


def get_service_version(application):
    try:
        serv_version = application.safe_data.get('workload-version', '')
    except Exception:
        serv_version = ''
    return serv_version


async def upgrade_services(model, upgraded, origin, origin_keys, upgrade_action, upgrade_params, pause, dry_run):
    sl_before = get_service_list(model, upgraded)
    log.info('Application upgrade order: {}'.format(
        ', '.join(['{} ({})'.format(name, version) for name, version in sl_before])
    ))

    s_upgrade = 0
    # upgrade_action is none by default, otherwise enforcing perform_upgrade
    use_action = upgrade_action if upgrade_action else 'openstack-upgrade'
    for app_name in upgraded:
        if app_name not in model.applications:
            continue
        rollable_app = await is_rollable(model.applications[app_name], use_action)
        if upgrade_action or rollable_app:
            await perform_upgrade(
                model,
                model.applications[app_name],
                origin_keys,
                use_action,
                upgrade_params,
                rollable=rollable_app,
                origin=origin,
                pause=pause,
                dry_run=dry_run
            )
            s_upgrade += 1
        else:
            await perform_bigbang_upgrade(
                model.applications[app_name],
                origin_keys,
                origin=origin,
                dry_run=dry_run
            )
            s_upgrade += 1

        await wait_until(
            model,
            model.applications.values(),
            timeout=1800,
            loop=model.loop
        )

    sl_after = get_service_list(model, upgraded)
    log.info('Application upgrade order: {}'.format(
        ', '.join(['{} ({}=>{})'.format(before[0], before[1], after[1]) for before, after in zip(sl_before, sl_after)])
    ))
    log.info('Upgrade finished ({} upgraded services)'.format(s_upgrade))


async def upgrade_charms(model, apps, dry_run, ignore_errors):
    """Upgrade charm revisions in the model.

    Listed apps will be checked for new revisions in charmstore
    and upgraded to latest revision if available.

    :param model_name: juju model name or uuid
    :param apps: ordered list of application names
    :param dry_run: boolean
    :param ignore_errors: boolean
    """
    log.info('Upgrading charms')
    upgraded = []
    latest_charms = []
    failed_upgrade = False

    for app_name in apps:
        attemp_update = False
        target_revision = None

        if app_name['charm'] not in model.applications:
            log.warning('Unable to find application: {}'.format(app_name['charm']))
            continue

        # Get charm url and parse current revision
        charm_url = model.applications[app_name['charm']].data['charm-url']
        parse = cs_name_parse(charm_url)
        current_revision = parse['revision'] if parse['revision'] else 0
        app = model.applications[app_name['charm']]

        # Charmstore get latest revision
        if parse.get('charmstore', False):
            try:
                if app_name.get('revision') is None:
                    charmstore_entity = await model.charmstore.entity(
                        parse['charm'], include_stats=False, includes=['revision-info']
                    )
                    # latest = charmstore_entity['Meta']['revision-info']['Revisions'][0]
                    charm_id = cs_name_parse(charmstore_entity['Id'])
                    target_revision = charm_id['revision']
                else:
                    target_revision = app_name['revision']
            except Exception:
                log.warning('Failed loading information from charmstore: {}'.format(charm_url))
                attemp_update = True
        else:
            log.info('Not upgrading local charm: {}'.format(charm_url))

        # Update if not newest or attempt to update if failed to find charmstore latest revision
        if (target_revision is not None and current_revision < target_revision) or attemp_update:
            log.info('Upgrade {} from: {} to: {}'.format(
                app_name['charm'], current_revision,
                'latest' if target_revision is None else target_revision
            ))
            try:
                if not dry_run:
                    await app.upgrade_charm(
                        revision=target_revision
                    )
                    await asyncio.sleep(30)
                upgraded.append(app_name['charm'])
            except JujuError:
                log.warning('Not upgrading: {}'.format(app_name['charm']))
            except Exception:
                log.error('Failed upgrading: {}'.format(app_name['charm']))
                if not ignore_errors:
                    failed_upgrade = True
                    break
        else:
            latest_charms.append(app_name['charm'])

    log.info('Upgraded: {} charms'.format(len(upgraded)))

    if not dry_run and upgraded:
        await asyncio.sleep(20)
        await wait_until(
            model,
            list(model.applications.values()),
            timeout=1800,
            loop=model.loop
        )

    log.info('Collecting final workload status')
    await asyncio.sleep(20)

    wss = Counter()
    wsm = Counter()
    for app in model.applications.values():
        if app.name in apps:
            wss += Counter([u.workload_status for u in app.units])
            wsm += Counter([
                u.workload_status_message for u in app.units if 'ready' not in u.workload_status_message
            ])
    log.info('Status of units after revision upgrade: {}'.format(dict(wss)))
    log.info('Workload messages: {}'.format(dict(wsm)))

    if not ignore_errors and (failed_upgrade or 'error' in wss.keys()):
        raise Exception('Errors during upgrade of charm revisions')
    return upgraded, latest_charms


async def wait_until(model, apps, log_time=10, timeout=None, wait_period=0.5, loop=None):
    """Blocking with logs.

    Return only after all conditions are true.
    Waiting for maintenance units to become active.

    :param model: juju model
    :param apps: list of juju applications
    :param log_time: logging frequency (s)
    :param timeout: blocking timeout (s)
    :param wait_period: waiting time between checks (s)
    :param loop: asyncio event loop
    """
    log_count = 0

    def _disconnected():
        return not (model.is_connected() and model.connection().is_open)

    async def _block(log_count):
        blockable = ['maintenance', 'blocked', 'waiting', 'error']
        while not _disconnected() and any(u.workload_status in blockable for a in apps for u in a.units):
            await asyncio.sleep(wait_period, loop=loop)
            log_count += 0.5
            if log_count % log_time == 0:
                wss = Counter([
                    unit.workload_status for a in apps for unit in a.units
                ])
                log.info('[WAITING] Charm workload status: {}'.format(dict(wss)))

    await asyncio.sleep(2)
    await asyncio.wait_for(_block(log_count), timeout, loop=loop)

    if _disconnected():
        raise websockets.ConnectionClosed(1006, 'no reason')


async def is_rollable(application, upgrade_action):
    """Define whether the application is rollable.

    Application is considered rollable if it provides upgrade action
    (e.g. openstack-upgrade), is deployed with more than 1 units,
    is not ceph and successfuly applies action-managed-upgrade config.

    :param application: juju application
    """
    actions = await enumerate_actions(application)

    if (not upgrade_action) or (upgrade_action not in actions):
        log.warn('Upgrade action "{}" not in actions.'.format(
            upgrade_action
        ))
        return False

    if len(application.units) <= 1:
        return False

    if application.name.lower().count('ceph') > 0:
        log.info('Ceph is not rollable.')
        return False

    if upgrade_action == 'openstack-upgrade' and not await application.set_config(
        {'action-managed-upgrade': 'True'}
    ):
        log.warn('Failed to enable action-managed-upgrade mode.')
        return False

    return True


def get_hacluster_subordinate_pairs(application):
    """Get hacluster subordinate pairs.

    Match hacluster subordinates into unit pairs, as they are respectively deployed.

    :param application: juju application
    """
    # TODO get app or unit? How are the relations described?
    # unit to unit relation missing in libjuju?!
    for relation in application.relations:
        if relation.is_subordinate and relation.provides.interface == 'hacluster':
            sub = relation.provides.application
            sub_pairs = {}
            # HACK (matus) only matching relation is IP address of unit and its subordinate
            # TODO machine id is not provided for subordinates in libjuju
            for unit in application.units:
                for sub_unit in sub.units:
                    if unit.data['public-address'] == sub_unit.data['public-address']:
                        sub_pairs[unit.name] = sub_unit
            return sub_pairs

    return {}


async def enumerate_actions(application):
    """Enumerate available actions for the applications.

    Returns a list of actions.

    :param application: juju application
    """
    actions = await application.get_actions()
    return actions.keys()


async def order_units(label, units):
    """Determing order of units.

    Returns a list of units beginning with leader.

    :param label: string
    :param units: list of juju units
    """
    log.info('{} - Determining order of units'.format(label))
    ordered = []

    is_leader_data = []
    for unit in units:
        is_leader_data.append(await unit.run('is-leader'))

    leader_info = filter(
        lambda u: u.data['results']['Stdout'].strip() == 'True',
        is_leader_data
    )
    leader_unit = [x.data['receiver'] for x in leader_info][0]

    for unit in units:
        if unit.name == leader_unit:
            ordered.insert(0, unit)
        else:
            ordered.append(unit)

    log.info('{} - Upgrade order is: {} (leader){}{}'.format(
        label,
        leader_unit,
        ', ' if len(ordered) > 1 else '',
        ', '.join([unit.name for unit in ordered if unit.name != leader_unit]),
    ))
    return ordered


async def perform_upgrade(
    model,
    application,
    origin_keys,
    upgrade_action,
    upgrade_params,
    dry_run=False,
    evacuate=False,
    rollable=False,
    pause=False,
    origin='cloud:xenial-ocata'
):
    """Perform upgrade.

    Rolling upgrade is performed on the rollable application.

    :param application: juju application
    :param dry_run: boolean
    :param evacuate: boolean
    :param rollable: boolean
    :param pause: boolean
    :param origin: origin string
    """
    label = application.name.upper()
    log.info('{} - Begin rolling upgrade'.format(label))

    actions = await enumerate_actions(application)
    if origin and not dry_run:
        config_key = origin_keys.get(application.name, 'openstack-origin')
        config = await application.get_config()
        previous = config.get(config_key, {}).get('value', '')

        if previous == origin:
            current = previous
        else:
            await application.set_config({config_key: origin})
            config_timeout = 60
            config_is_set = False
            while config_timeout > 0 and not config_is_set:
                config_timeout = config_timeout - 1
                await asyncio.sleep(5)
                config = await application.get_config()
                current = config.get(config_key, {}).get('value', '')
                config_is_set = current == origin
                log.info('{} - Setting config {} = {} => {}'.format(label, config_key, previous, current))
        log.info('{} - Config {} = {}'.format(label, config_key, current))

    ordered_units = await order_units(label, application.units) if rollable else application.units
    hacluster_pairs = get_hacluster_subordinate_pairs(application) if (rollable and pause) else {}  # TODO see fx

    for unit in ordered_units:
        hacluster_unit = hacluster_pairs.get(unit.name, False)

        if evacuate and application.name == 'nova-compute':
            # NOT IMPLEMENTED
            log.warn('{} - Nova evacuation is not implemented, app will be skipped'.format(label))
            break

        if pause and hacluster_unit:
            # TODO this will pause all the units for hacluster subordinates
            log.info('{} - Pausing service on hacluster subordinate: {}'.format(label, hacluster_unit.name))
            if not dry_run:
                async with async_timeout.timeout(300):
                    action = await hacluster_unit.run_action('pause')
                    await action.wait()
                    log.debug('{} - Service action: {} on unit: {} status: {}'.format(
                        label, unit.name, 'pause', action.status
                    ))
            log.info('{} - Service on hacluster subordinate {} is paused'.format(label, hacluster_unit.name))

        if pause and 'pause' in actions:
            log.info('{} - Pausing service on unit: {}'.format(label, unit.name))
            if not dry_run:
                async with async_timeout.timeout(300):
                    action = await unit.run_action('pause')
                    await action.wait()
                    log.debug('{} - Service action: {} on unit: {} status: {}'.format(
                        label, unit.name, 'pause', action.status
                    ))
            log.info('{} - Service on unit {} is paused'.format(label, unit.name))

        if upgrade_action in actions:
            log.info('{} - Upgrading service for unit: {}'.format(label, unit.name))
            if not dry_run:
                action = await unit.run_action(upgrade_action, **upgrade_params)
                await action.wait()
                log.debug('{} - Service action: {} on unit: {} status: {}'.format(
                    label, unit.name, upgrade_action, action.status
                ))
            log.info('{} - Completed upgrade for unit: {}'.format(label, unit.name))

        if pause and 'resume' in actions:
            log.info('{} - Resuming service on unit: {}'.format(label, unit.name))
            if not dry_run:
                async with async_timeout.timeout(300):
                    action = await unit.run_action('resume')
                    await action.wait()
                    log.debug('{} - Service action: {} on unit: {} status: {}'.format(
                        label, unit.name, 'resume', action.status
                    ))
            log.info('{} - Service on unit {} has resumed'.format(label, unit.name))

        if pause and hacluster_unit:
            # TODO this will resume all the units for hacluster subordinates
            log.info('{} - Resuming service on hacluster subordinate: {}'.format(label, hacluster_unit.name))
            if not dry_run:
                async with async_timeout.timeout(300):
                    action = await hacluster_unit.run_action('resume')
                    await action.wait()
                    log.debug('{} - Service action: {} on unit: {} status: {}'.format(
                        label, unit.name, 'resume', action.status
                    ))
            log.info('{} - Service on hacluster subordinate {} has resumed'.format(label, hacluster_unit.name))

        await wait_until(
            model,
            list(model.applications.values()),
            timeout=1800,
            loop=model.loop
        )
        log.info('{} - Unit {} has finished the upgrade'.format(label, unit.name))
    log.info('{} - Finish rolling upgrade'.format(label))


async def perform_bigbang_upgrade(
    application,
    origin_keys,
    dry_run=False,
    pause=False,
    origin='cloud:xenial-ocata'
):
    """Perform bigbang upgrade.

    Bigbang upgrade is performed on units of the application at once.

    :param application: juju application
    :param dry_run: boolean
    :param pause: boolean
    :param origin: origin string
    """
    log.info('Big-bang upgrade: {}'.format(application.name))
    if origin and not dry_run:
        config_key = origin_keys.get(application.name, 'openstack-origin')
        if config_key not in await application.get_config():
            log.warn('Unable to set source/origin during big-bang upgrade for service: {}'.format(application.name))
            log.info('Skipping upgrade for service: {}'.format(application.name))
            return

        log.info('Set {} config {}={}'.format(application.name, config_key, origin))
        await application.set_config({config_key: origin})
        await asyncio.sleep(15)

    if not dry_run:
        upgrade_in_progress = True
        while upgrade_in_progress:
            if any(
                [u.workload_status.lower().find('upgrad') >= 0 for u in application.units]
            ):
                await asyncio.sleep(5)
            else:
                upgrade_in_progress = False



from pathlib import Path
from setuptools import setup, find_packages

here = Path(__file__).absolute().parent
readme = here / 'README.md'
version = here / 'VERSION'

SETUP = {
    'name': "jujuna",
    'description': 'Continuous deployment, upgrade and testing for Juju.',
    'packages': find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    'version': version.read_text().strip(),
    'author': "Matus Kosut",
    'author_email': "matuskosut@gmail.com",
    'maintainer': 'HUNT Data Center',
    'maintainer_email': 'cloud@hunt.ntnu.no',
    'url': "https://github.com/huntdatacenter/jujuna",
    'long_description': readme.read_text().strip(),
    'long_description_content_type': 'text/markdown',
    'entry_points': {
        'console_scripts': [
            # Script invokation:
            'jujuna = jujuna.__main__:main',
        ]
    },
    'license': 'Apache 2',
    'classifiers': [
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Environment :: Console"
    ],
    'install_requires': [
        'async-timeout<4.0.0,>=2.0.1',
        'argcomplete==1.10.0',
        'theblues<1.0.0,>=0.5.2'
    ],
    'extras_require': {
        ":python_version>'3.5.2'": [
            'juju<2.9.0,>=2.8.0',
            'pyyaml<6.0,>=3.0',
        ],
        ":python_version<='3.5.2'": [
            'juju<1.0.0,>=0.11.7',
            'pyyaml<=4.2,>=3.0',
        ]
    }
}


if __name__ == '__main__':
    setup(**SETUP)



"""Network broker."""
from . import Broker, python3, load_output
from jujuna.exporters import Exporter
import logging


log = logging.getLogger('jujuna.tests.broker')


class Network(Broker):
    """Network broker."""

    def __init__(self):
        """Init broker."""
        super().__init__()

    async def run(self, test_case, unit, idx):
        """Run tests."""
        rows = []
        async with Exporter(unit, self.named) as exporter:
            if 'port' in test_case:
                try:
                    test_data = await unit.run(python3(exporter), timeout=10)
                    results = load_output(test_data.data['results'])
                except Exception as exc:
                    log.debug(exc)
                    results = {'sockets': []}
                # print(results)

                local_ports = [str(s['local_port']) for s in results['sockets']]

                for port, open in test_case['port'].items():
                    status = 'open' if open else 'closed'
                    rows.append((idx, '{}.{} == {}'.format('port', port, status), (str(port) in local_ports) == open), )

        return rows



"""Broker abstract class."""
import json


class Broker():
    """Broker abstract class."""

    def __init__(self):
        """Init abstract class."""
        self.named = self.__class__.__name__.lower()

    async def run(self, *args, **kwargs):
        """Run method have to be overriden."""
        print("Run method for '{}' not implemented".format(self.named))
        return []


def python3(file, args=[]):
    if args:
        return 'python3 {} {}'.format(file, ' '.join([str(arg) for arg in args]))
    else:
        return 'python3 {}'.format(file)


def load_output(results):
    if results['Code'] == '0':
        try:
            var = json.loads(results['Stdout'])
        except Exception as e:
            raise Exception('JSON load failed: {}'.format(e))
    else:
        raise Exception('Operation failed: {}'.format(results['Stderr']))
    return var
